﻿#-------------------------------------------------------------------------------
# Name:        RINEX
# Purpose:      RINEXファイルを取り扱うクラス
#               とりあえずの開発目標は、RINEXファイルの結合です。
#
# Author:      morishita
#
# Created:     02/02/2012
# Copyright:   (c) morishita 2012
# Licence:     new BSD
#-------------------------------------------------------------------------------
#!/usr/bin/env python

import os
import re
import sys

def isRINEX(fname):
    """ RINEXファイルかどうかを判定する
    判定結果はTrue or False
    拡張子も判断材料とする
    """
    if os.path.exists(fname) == False:      # ファイルの存在が確認できない場合はFalseを返す
        return False
    if os.path.isfile(fname) != True:       # ファイルでなければFalseを返す
        return False
    root, ext = os.path.splitext(fname)     # 拡張子を取得
    if re.search(".\d+[onmgh]", fname) != None:# 正規表現を利用して拡張子のチェック
        fr = open(fname, 'r')               # ファイルを開く
        firstLine = fr.readline()           # 1行だけ取得
        fr.close()
        if "RINEX" in firstLine:            # 1行目に“RINEX”とあればOK
            return True
        else:
            return False
    else:
        return False

class Epoch:
    """ エポックを表すクラス
    その内、GPS時刻にも対応させたいけどうるう秒の挿入日リストを整備しないといけないなぁ。。。
    """
    def __init__(self, field = ""):
        self.year = 0
        self.month = 0
        self.day = 0
        self.hour = 0
        self.min = 0
        self.sec = 0.0
        if isinstance(field, list):
            self.set(field)
        return
    def set(self, field):
        """ 渡された文字列でエポックを初期化する
        エラー処理は一切ないので注意
        """
        self.year = int(field[0])
        self.month = int(field[1])
        self.day = int(field[2])
        self.hour = int(field[3])
        self.min = int(field[4])
        self.sec = float(field[5])
        return
    def isSmallerThan(self, epoch):
        """ 引数のエポックと比較して小さければTrueを返す
        """
        if self.year < epoch.year:
            return True
        elif self.month < epoch.month:
            return True
        elif self.day < epoch.day:
            return True
        elif self.hour < epoch.hour:
            return True
        elif self.min < epoch.min:
            return True
        elif self.sec < epoch.sec:
            return True
        else:
            return False


class HeaderOfRINEX:
    """ RINEXヘッダ情報
    """
    def __init__(self, rinexFile = ""):
        """ RINEXファイルよりヘッダ情報を取得する
        """
        self.header = []
        self.start = ""
        self.end = ""
        self.isSet = False                              # セットされるとTrueになる
        self.set(rinexFile)                             # もし可能ならヘッダー情報をセットする
        return
    def set(self, fname = ""):
        """ 指定されたファイルのヘッダー情報をセットする
        上書きされるので注意
        """
        if isRINEX(fname):
            self.header = []
            fr = open(fname,'r')                        # ファイルを開く
            txt = fr.readlines()
            fr.close()
            for line in txt:
                if "start" in line:
                    self.start = line
                elif "end" in line:
                    self.end = line
                self.header.append(line)
                self.isSet = True
                if "END OF HEADER" in line:             # ヘッダー部分を抜けるとループを停止
                    break
        return
    def fusion(self, fname):
        """ ファイル名を指定して、含まれるヘッダー情報から終了時刻情報を抜き出して時刻を上書きする
        日本語が変かも。
        """
        if isRINEX(fname) and self.isSet == True:       # 一度はセットしておかなければ
            fr = open(fname,'r')                        # ファイルを開く
            txt = fr.readlines()
            fr.close()
            for line in txt:
                if "end" in line:
                    self.end = line
                if "END OF HEADER" in line:             # ヘッダー部分を抜けるとループを停止
                    break
        return
    def getHeader(self):
        """ ヘッダ情報をリストとして返す
        """
        ans = []
        for line in self.header:
            if "end" in line:                           # end（終了時刻）だけは置換する
                ans.append(self.end)
            else:
                ans.append(line)
        return ans

class RINEX:
    def __init__(self):
        self.txt = []                                       # リストを初期化
        self.header = HeaderOfRINEX()
        self.lastEpoch = Epoch()
        self.epochPattern = re.compile(r"\d+ +\d+ +\d+ +\d+ +\d+ +\d+.\d+ +.+")  # エポックに一致する正規表現
        self.delimiter = re.compile(r" +")                  # 半角スペースがデリミタになっている場合に使用する正規表現。複数のスペースにマッチする。
    def join(self, fname):
        """ 指定されたRINEXファイルを結合します
        """
        if os.path.exists(fname) == False:                  # ファイルの存在が確認できない場合は素早く抜ける
            return
        if isRINEX(fname):
            _header = HeaderOfRINEX(fname)                  # ヘッダー情報を取得
            if self.header.isSet == False:                  # 未セットならセットする
                self.header = _header
            else:
                self.header.fusion(fname)                   # ヘッダー情報を統合
            fr = open(fname,'r')                            # ファイルを開く
            _txt = fr.readlines()
            fr.close()
            _headerRead = False
            _copyEnable = False
            for line in _txt:
                if _headerRead:                             # ヘッダー部分を読み飛ばしていれば以下を処理する
                    _line = line.lstrip(" ").rstrip()       # 両端から空白と改行文字を削除
                    #print(_line)
                    _machOfEpoch = self.epochPattern.search(_line)  # 正規表現を使って、エポックを表す行であるか検査
                    #result3 = re.findall("G *\d+", _line)  # 衛星番号にヒットする
                    #print(result3)
                    if _machOfEpoch != None:                # マッチすればNoneではない
                        #print("epoch.")
                        field = self.delimiter.split(_line) # 半角スペース(複数の半角スペースでもOK)で分割
                        #print(field)
                        _epoch = Epoch(field)               # 文字列解析してエポックを取得
                        if self.lastEpoch.isSmallerThan(_epoch):# 保持しているエポックよりも大きいことを確認
                            self.lastEpoch = _epoch
                            _copyEnable = True
                        else:
                            _copyEnable = False             # 古いエポックならコピーしない
                    if _copyEnable:
                        self.txt.append(line)               # 文字列を結合させる
                if "END OF HEADER" in line:                 # ヘッダー部分を読み飛ばす仕掛け
                    _headerRead = True
    def save(self, saveName):
        """ 指定されたファイル名で保存する
        """
        fw = open(saveName,'w')                             # 書き込み用にファイルを開く
        _header = self.header.getHeader()                   # ヘッダー情報を取得
        for line in _header:
            fw.write(line)                                  # ヘッダー情報をファイルへ書き込む
        for line in self.txt:
            fw.write(line)                                  # ファイルへ書き込む
        fw.close()
        return


def main():
    """ デフォルトの実行メソッドですセルフテストに使用します
    引数にRINEXファイル名を引き渡すと処理するはずです。
    """
    print("Program start...")
    print("This program combine some RINEX files.\n")
    argvs = sys.argv                                        # 引数を取得する
    argc = len(argvs)
    if argc > 1:
        _rinex = RINEX()
        for i in range(1, argc):
            fname = argvs[i]
            if isRINEX(fname):                              # RINEXファイルかどうかチェックする
                print("now target: " + fname)
                _rinex.join(fname)                          # 追加する（実行が時系列に行われることが前提）
            else:
                print("This file is not RINEX file.")
        _rinex.save("combined.txt")
    else:
        print("argv is zero. Input RINEX file names.")
    print("\nThe combined text was created.")
    print("Program fin.")

if __name__ == '__main__':
    main()
